{-
  PP, laboratorul 8: tipuri de date utilizator
-}

import Data.List
import TestPP

{-
  1. (2p) Numere complexe
  Se dă tipul de date Complex, reprezentând numerele complexe, i.e.
  primul câmp îl reprezintă partea reală, iar cel de-al doilea partea
  imaginară.

  Implementați următoarele operații cu numere complexe:
  - compararea la egalitate (eqC) între două numere
  - adunarea (addC) și scăderea (subC) a două numere
  - conjugata (conjC) unui număr
  - modulul (absC) unui număr

  Pentru mai multe detalii, consultați:
  https://en.wikipedia.org/wiki/Complex_number#Elementary_operations
-}
test1 :: TestPP ()
test1 = tests 1 2
        [ testCond (eqC c1 c1) "eqC" 0.5
	, testCond (((c1 `addC` c2) `subC` (C 1 1)) `eqC` (C 6 6)) "addC, subC" 0.5
	, testCond ((conjC c2) `eqC` (C 4 (-5))) "conjC" 0.5
	, testCond ((absC c2) `eqR` 6.4031) "absC" 0.5
	]
    where
          c1 = C 3 2
          c2 = C 4 5
	  eqR a b = abs (a - b) < 0.0001


data Complex = C
  { realC :: Double
  , imgC  :: Double
  } deriving Show

-- Hint: dat fiind că numerele complexe sunt definite pe baza a două
-- numere în virgulă mobilă, puteți implementa egalitatea aproximativă
-- dintre două numere, folosind absC și subC.
eqC :: Complex -> Complex -> Bool
eqC c1 c2 = (absC $ c1 `subC` c2) < eps
  where eps = 0.0001

addC :: Complex -> Complex -> Complex
addC (C a b) (C c d) = C (a + c) (b + d)

subC :: Complex -> Complex -> Complex
subC (C a b) (C c d) = C (a - c) (b - d)

conjC :: Complex -> Complex
conjC (C a b) = C a (-b)

absC :: Complex -> Double
absC (C a b) = sqrt $ a * a + b * b



{-
 2. (4p) Liste imbricate
  Se dă tipul de date SList a care are funcționalități
  asemănătoare listelor din limbajele Lisp (e.g. Scheme, Racket, Clojure),
  permițând componente la diferite niveluri de imbricare.
  Ex: Lista din Racket '(1 (3 4) (2)) să poată fi definită în Haskell
  folosind SList.
  Definiți:
  - emptySList, lista vidă
  - consElem, adaugă un element în capul unei liste
    Ex: consElem 1 '((3 4) (2)) == '(1 (3 4) (2))
  - consList, adaugă o listă (imbricată) în capul unei liste
    Ex: consList '(2 3) '(1 2) == '((2 3) 1 2)
  - headSList, ia primul element dintr-un SList
  - tailSList, ia restul SList-ului
  - deepEqual, o funcție ce verifică egalitatea a două SList - este
    interzisă rezolvarea prin derivarea clasei Eq
  - flatten, întoarce lista cu elementele din SList (pe același nivel)
-}

test2 :: TestPP ()
test2 = tests 2 4
        [ testCond (deepEqual l1 l1 && not (deepEqual l1 l2)) "simple lists" 0.5
	, testCond (deepEqual (consElem 2 $ l3) l2) "less simple lists" 0.75
	, testCond (not (deepEqual (consList l1 emptySList) l4)) "different lengths" 0.5
	, testCond (deepEqual (consList l1 (Atom 1)) l4) "consList" 0.5
	, testCond (deepEqual (headSList $ tailSList l2)
	                      (consElem 1 $ consElem 1 emptySList)) "head, tail" 0.5
	, testCond (deepEqual (tailSList $ consList l1 (Atom 1)) l1) "consList tail" 0.5
	, testVal (flatten l2) [2, 1, 1, 3] "flatten" 0.75
	]
    where
          l1 = consElem 1 $ emptySList
          l2 = consElem 2 $ consList (consElem 1 $ consElem 1 emptySList) $
               consElem 3 emptySList
          l3 = consList (consElem 1 $ consElem 1 emptySList) $ consElem 3 $
               emptySList
	  l4 = consList l1 $ consElem 1 $ emptySList



data SList a = Atom a | List [SList a] deriving Show

emptySList :: SList a
emptySList = List []

consElem :: a -> SList a -> SList a
consElem x (Atom y)  = List [Atom x, Atom y]
consElem x (List xs) = List (Atom x : xs)

consList :: SList a -> SList a -> SList a
consList x (Atom y)  = List [x, Atom y]
consList x (List ys) = List $ x : ys

headSList :: SList a -> SList a
headSList (List x) = head x
headSList _        = error "head undefined on Atoms"

tailSList :: SList a -> SList a
tailSList (List x) = List $ tail x
tailSList _        = error "tail undefined on Atoms"

deepEqual :: Eq a => SList a -> SList a -> Bool
deepEqual (Atom x) (Atom y) = x == y
deepEqual (List a) (List b)
      | length a /= length b = False
      | otherwise = and $ zipWith deepEqual a b
deepEqual _ _               = False

flatten :: SList a -> [a]
flatten (Atom x)  = [x]
flatten (List xs) = concatMap flatten xs


{-
  3. (4p) Arbori binari de căutare
  Definiți funcții pentru a crea un arbore binar de căutare pornind de la o
  listă de elemente, căutarea unui element într-un arbore binar de căutare
  dat și o funcție care întoarce lista elementelor din parcurgerea în
  inordine a arborelui.

  Hint: Este de preferat ca arborele binar de căutare să fie echilibrat,
  lucru ușor de obținut la creare dacă lista de elemente este sortată.
-}


test3 :: TestPP ()
test3 = tests 3 4
        [ testVal (findElem 1 bst) (Just 1) "findElem" 2
	, testVal (inorder bst) [1, 2, 3, 4] "inorder" 2
	]
    where
         bst = makeBST [2, 3, 1, 4]


data BST a = BSTNod a (BST a) (BST a) | BSTNil deriving Show

makeBST :: Ord a => [a] -> BST a
makeBST lst = makeBSTHelper (sort lst)

makeBSTHelper [] = BSTNil
makeBSTHelper lst = BSTNod m (makeBSTHelper st) (makeBSTHelper dr)
  where
  l = length lst
  m = lst !! (div l 2)
  st = take (div l 2) lst
  dr = drop ((div l 2) + 1) lst

findElem :: (Ord a, Eq a) => a -> BST a -> Maybe a
findElem _ BSTNil = Nothing
findElem y (BSTNod x arbSt arbDr)
      | x == y = Just x
      | y < x = findElem y arbSt
      | otherwise = findElem y arbDr

inorder :: BST a -> [a]
inorder BSTNil                 = []
inorder (BSTNod x arbSt arbDr) = inorder arbSt ++ [x] ++ inorder arbDr

{-
  4. (BONUS, 1.5p) Arbori binari de căutare
  Lucrând pe același tip de date de la exercițiul anterior, definiți
  funcția care află strămoșul comun a două noduri.
  Vom lucra doar pe arbori ce nu conțin valori duplicate!
-}
test4 :: TestPP ()
test4 = tests 4 1.5
        [ testVal (lowestCommonAncestor t 1 3) (Just 2) "lowestCommonAncestor 1, 3" 0.5
        , testVal (lowestCommonAncestor t 4 5) (Just 4) "lowestCommonAncestor 4, 5" 0.5
	, testVal (lowestCommonAncestor t 4 7) Nothing "lowestCommonAncestor 4, 7" 0.5
	]
    where
          t = makeBST [1, 2, 3, 4, 5, 6]

lowestCommonAncestor :: (Ord a, Eq a) => BST a -> a -> a -> Maybe a
lowestCommonAncestor t n1 n2 = if findElem n1 t == Nothing || findElem n2 t == Nothing
					then Nothing
					else lcaHelper t n1 n2
        where
            lcaHelper (BSTNod x arbSt arbDr) n1 n2
                | n1 == x || n2 == x = Just x
                | n1 < x && n2 < x = lowestCommonAncestor arbSt n1 n2
                | n1 > x && n2 > x = lowestCommonAncestor arbDr n1 n2
                | otherwise = Just x


{-
  5. (BONUS, 1.5p) Arbori binari de căutare
  Definiți funcția nodesDistance, care returnează distanța
  între două noduri (numărul minim de arce ce trebuie parcurse
  pentru a ajunge dintr-un nod în celălalt).
  Vom lucra doar pe arbori ce nu conțin valori duplicate!
-}
test5 :: TestPP ()
test5 = tests 5 1.5
        [ testVal (nodesDistance t 1 5) (Just 4) "nodesDistance 1, 5" 0.5
        , testVal (nodesDistance t 4 5) (Just 2) "nodesDistance 4, 5" 0.5
	, testVal (nodesDistance t 2 7) Nothing "nodesDistance 2, 7" 0.5
	]
    where
          t = makeBST [1, 2, 3, 4, 5, 6]

nodesDistance :: (Ord a, Eq a) => BST a -> a -> a-> Maybe Int
nodesDistance t n1 n2 = case lowestCommonAncestor t n1 n2 of
	Nothing -> Nothing
	(Just lca) -> let r_n1 = distance t n1
            		  r_n2 = distance t n2
                          r_lca = distance t lca
		      in Just $ r_n1 + r_n2 - 2 * r_lca
        where
            distance (BSTNod x arbSt arbDr) n
                | n == x = 0
                | n < x = 1 + distance arbSt n
                | otherwise = 1 + distance arbDr n

{-
  6. (BONUS, 2p) Arbori binari de căutare
  Definiți funcția subtree, care verifică dacă
  arborele t1 este un subarbore al arborelui t2.
  Este interzisă rezolvarea prin derivarea clasei Eq!
  Ex: subtree (makeBST [1,2]) (makeBST [1,2,3,4]) == True
  Vom lucra doar pe arbori ce nu conțin valori duplicate!
-}


test6 :: TestPP ()
test6 = tests 6 2
        [ testCond (subtree (BSTNod 2 (BSTNod 1 BSTNil BSTNil) BSTNil) t) "subtree [1, 2]" 0.75
	, testCond (subtree t t) "subtree [1, 2, 3, 4]" 0.5
	, testCond (not $ subtree (BSTNod 4 (BSTNod 3 BSTNil BSTNil) BSTNil) t) "not $ subtree [3, 4]" 0.75
	]
    where
          t = BSTNod 3 (BSTNod 2 (BSTNod 1 BSTNil BSTNil) BSTNil) (BSTNod 4 BSTNil BSTNil)


subtree :: (Eq a, Ord a) => BST a -> BST a -> Bool
subtree BSTNil _ = True
subtree _ BSTNil = False
subtree s@(BSTNod x1 _ _) t@(BSTNod x2 arbSt arbDr)
       | eqTree s t = True
       | x1 < x2 = subtree s arbSt
       | x1 > x2 = subtree s arbDr
       | otherwise = False

eqTree :: Eq a => BST a -> BST a -> Bool
eqTree BSTNil BSTNil = True
eqTree BSTNil _ = False
eqTree _ BSTNil = False
eqTree (BSTNod x1 arbSt1 arbDr1) (BSTNod x2 arbSt2 arbDr2)
       | x1 == x2 = eqTree arbSt1 arbSt2 && eqTree arbDr1 arbDr2
       | otherwise = False


{-
Helpers for testing :)
-}
allTests = [test1, test2, test3, test4, test5, test6]
check = runTestPP $ sequence_ allTests
[check1, check2, check3, check4, check5, check6] = map runTestPP allTests
